/*
 * flexcop.c - driver for B2C2 FlexcopII(b)/III Digital TV chip
 *
 * Copyright (C) 2003  Vadim Catana, skystar@moldova.cc
 *
 * FIX: DISEQC Tone Burst in flexcop_diseqc_ioctl()
 * FIX: FULL soft DiSEqC for skystar2 (FlexCopII rev 130) VP310 equipped
 *     Vincenzo Di Massa, hawk.it at tiscalinet.it
 *
 * Converted to Linux coding style
 * Misc reorganization, polishing, restyling
 *     Roberto Ragusa, r.ragusa at libero.it
 *
 * Added hardware filtering support,
 *     Niklas Peinecke, peinecke at gdv.uni-hannover.de
 * 
 * Added support for USB devices
 *     Bjarne Steinsbo, bjarne at steinsbo.com
 *     Patrick Boettcher, patrick.boettcher at desy.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "flexcop.h"

int b2c2_flexcop_debug;
module_param_named(debug, b2c2_flexcop_debug,  int, 0644);
MODULE_PARM_DESC(debug, "set debugging level (1=info,2=i2c,4=tuner (|-able)).");


#if 0
void b2c2_write_reg_bitfield(struct b2c2_device *card, u16 reg, u32 zeromask, u32 orvalue)
{
	u32 tmp;

	tmp = card->read_dw_old (card, reg);
	tmp = (tmp & ~zeromask) | orvalue;
	card->write_dw_old (card, reg, tmp);
}

/* SRAM (Skystar2 rev2.3 has one "ISSI IS61LV256" chip on board,
   but it seems that FlexCopII can work with more than one chip) */
void b2c2_sram_set_net_dest(struct b2c2_device *card, u8 dest)
{
	u32 tmp;

	udelay(1000);

	tmp = (card->read_dw_old(card, 0x714) & 0xfffffffc) | (dest & 3);

	udelay(1000);

	card->write_dw_old(card, 0x714, tmp);
	card->write_dw_old(card, 0x714, tmp);

	udelay(1000);

	/*return value is never used? */
/*      return tmp; */
}

void b2c2_sram_set_cai_dest(struct b2c2_device *card, u8 dest)
{
	u32 tmp;

	tmp = (card->read_dw_old(card, 0x714) & 0xfffffff3) | ((dest & 3) << 2);

	udelay(1000);
	udelay(1000);

	card->write_dw_old(card, 0x714, tmp);
	card->write_dw_old(card, 0x714, tmp);

	udelay(1000);

	/*return value is never used? */
/*      return tmp; */
}

void b2c2_sram_set_cao_dest(struct b2c2_device *card, u8 dest)
{
	u32 tmp;

	tmp = (card->read_dw_old(card, 0x714) & 0xffffffcf) | ((dest & 3) << 4);

	udelay(1000);
	udelay(1000);

	card->write_dw_old(card, 0x714, tmp);
	card->write_dw_old(card, 0x714, tmp);

	udelay(1000);

	/*return value is never used? */
/*      return tmp; */
}

void b2c2_sram_set_media_dest(struct b2c2_device *card, u8 dest)
{
	u32 tmp;

	udelay(1000);

	tmp = (card->read_dw_old(card, 0x714) & 0xffffff3f) | ((dest & 3) << 6);

	udelay(1000);
	udelay(1000);

	card->write_dw_old(card, 0x714, tmp);
	card->write_dw_old(card, 0x714, tmp);

	udelay(1000);
	
	/* return value is never used? */
/*      return tmp; */
}
/* SRAM memory is accessed through a buffer register in the FlexCop
   chip (0x700). This register has the following structure:
    bits 0-14  : address
    bit  15    : read/write flag
    bits 16-23 : 8-bit word to write
    bits 24-27 : = 4
    bits 28-29 : memory bank selector
    bit  31    : busy flag
*/
static void flex_sram_write(struct b2c2_device *card, u32 bank, u32 addr, u8 *buf, u32 len)
{
	int i, retries;
	u32 command;

	for (i = 0; i < len; i++) {
		command = bank | addr | 0x04000000 | (*buf << 0x10);

		retries = 2;

		while (((card->read_dw_old(card, 0x700) & 0x80000000) != 0) && (retries > 0)) {
			mdelay(1);
			retries--;
		}

		if (retries == 0)
			b2c2_err("SRAM timeout");

		card->write_dw_old(card, 0x700, command);

		buf++;
		addr++;
	}
}

static void flex_sram_read(struct b2c2_device *card, u32 bank, u32 addr, u8 *buf, u32 len)
{
	int i, retries;
	u32 command, value;

	for (i = 0; i < len; i++) {
		command = bank | addr | 0x04008000;

		retries = 10000;

		while (((card->read_dw_old(card, 0x700) & 0x80000000) != 0) && (retries > 0)) {
			mdelay(1);
			retries--;
		}

		if (retries == 0)
			b2c2_err("SRAM timeout");

		card->write_dw_old(card, 0x700, command);

		retries = 10000;

		while (((card->read_dw_old(card, 0x700) & 0x80000000) != 0) && (retries > 0)) {
			mdelay(1);
			retries--;
		}

		if (retries == 0)
			b2c2_err("SRAM timeout");

		value = card->read_dw_old(card, 0x700) >> 0x10;

		*buf = (value & 0xff);

		addr++;
		buf++;
	}
}

static void sram_write_chunk(struct b2c2_device *card, u32 addr, u8 *buf, u16 len)
{
	u32 bank;

	bank = 0;

	if (card->dw_sram_type == 0x20000) {
		bank = (addr & 0x18000) << 0x0d;
	}

	if (card->dw_sram_type == 0x00000) {
		if ((addr >> 0x0f) == 0)
			bank = 0x20000000;
		else
			bank = 0x10000000;
	}

	 flex_sram_write(card, bank, addr & 0x7fff, buf, len);
}

static void sram_read_chunk(struct b2c2_device *card, u32 addr, u8 *buf, u16 len)
{
	u32 bank;

	bank = 0;

	if (card->dw_sram_type == 0x20000) {
		bank = (addr & 0x18000) << 0x0d;
	}

	if (card->dw_sram_type == 0x00000) {
		if ((addr >> 0x0f) == 0)
			bank = 0x20000000;
		else
			bank = 0x10000000;
	}

	 flex_sram_read(card, bank, addr & 0x7fff, buf, len);
}

static void sram_read(struct b2c2_device *card, u32 addr, u8 *buf, u32 len)
{
	u32 length;

	while (len != 0) {
		length = len;

		/* check if the address range belongs to the same
		   32K memory chip. If not, the data is read from
		   one chip at a time. */
		if ((addr >> 0x0f) != ((addr + len - 1) >> 0x0f)) {
			length = (((addr >> 0x0f) + 1) << 0x0f) - addr;
		}

		sram_read_chunk(card, addr, buf, length);

		addr = addr + length;
		buf = buf + length;
		len = len - length;
	}
}

static void sram_write(struct b2c2_device *card, u32 addr, u8 *buf, u32 len)
{
	u32 length;

	while (len != 0) {
		length = len;

		/* check if the address range belongs to the same
		   32K memory chip. If not, the data is read from
		   one chip at a time. */
		if ((addr >> 0x0f) != ((addr + len - 1) >> 0x0f)) {
			length = (((addr >> 0x0f) + 1) << 0x0f) - addr;
		}

		sram_write_chunk(card, addr, buf, length);

		addr = addr + length;
		buf = buf + length;
		len = len - length;
	}
}

static void sram_set_size(struct b2c2_device *card, u32 mask)
{
	card->write_dw_old(card, 0x71c, (mask | (~0x30000 & card->read_dw_old(card, 0x71c))));
}

static void sram_init(struct b2c2_device *card)
{
	u32 tmp;

	tmp = card->read_dw_old(card, 0x71c);

	card->write_dw_old(card, 0x71c, 1);

	if (card->read_dw_old(card, 0x71c) != 0) {
		card->write_dw_old(card, 0x71c, tmp);

		card->dw_sram_type = tmp & 0x30000;
	} else {
		card->dw_sram_type = 0x10000;
	}
	b2c2_info("dw_sram_type = %x",card->dw_sram_type);

	/* return value is never used? */
/*      return adapter->dw_sram_type; */
}

static int sram_test_location(struct b2c2_device *card, u32 mask, u32 addr)
{
	u8 tmp1, tmp2;

	b2c2_dinfo(card, "mask = %x, addr = %x",mask, addr);

	sram_set_size(card, mask);
	sram_init(card);

	tmp2 = 0xa5;
	tmp1 = 0x4f;

	sram_write(card, addr, &tmp2, 1);
	sram_write(card, addr + 4, &tmp1, 1);

	tmp2 = 0;

	mdelay(20);

	sram_read(card, addr, &tmp2, 1);
	sram_read(card, addr, &tmp2, 1);

	b2c2_dinfo(card, "wrote 0xa5, read 0x%2x",tmp2);

	if (tmp2 != 0xa5)
		return 0;

	tmp2 = 0x5a;
	tmp1 = 0xf4;

	sram_write(card, addr, &tmp2, 1);
	sram_write(card, addr + 4, &tmp1, 1);

	tmp2 = 0;

	mdelay(20);

	sram_read(card, addr, &tmp2, 1);
	sram_read(card, addr, &tmp2, 1);

	b2c2_dinfo(card, "wrote 0x5a, read 0x%2x",tmp2);

	if (tmp2 != 0x5a)
		return 0;

	return 1;
}

static u32 sram_length(struct b2c2_device *card)
{
	if (card->dw_sram_type == 0x10000)
		return 32768;	/* 32K */
	if (card->dw_sram_type == 0x00000)
		return 65536;	/* 64K */
	if (card->dw_sram_type == 0x20000)
		return 131072;	/* 128K */

	return 32768;	/* 32K */
}

/* FlexcopII can work with 32K, 64K or 128K of external SRAM memory.
    - for 128K there are 4x32K chips at bank 0,1,2,3.
    - for  64K there are 2x32K chips at bank 1,2.
    - for  32K there is one 32K chip at bank 0.

   FlexCop works only with one bank at a time. The bank is selected
   by bits 28-29 of the 0x700 register.

   bank 0 covers addresses 0x00000-0x07fff
   bank 1 covers addresses 0x08000-0x0ffff
   bank 2 covers addresses 0x10000-0x17fff
   bank 3 covers addresses 0x18000-0x1ffff
*/
static int sram_detect_for_flex2(struct b2c2_device *card)
{
	u32 tmp, tmp2, tmp3;

	tmp = card->read_dw_old(card, 0x208);
	card->write_dw_old(card, 0x208, 0);

	tmp2 = card->read_dw_old(card, 0x71c);

	b2c2_dinfo(card, "tmp2 = %x",tmp2);

	card->write_dw_old(card, 0x71c, 1);

	tmp3 = card->read_dw_old(card, 0x71c);

	b2c2_dinfo(card, "tmp3 = %x",tmp3);

	card->write_dw_old(card, 0x71c, tmp2);

	/* check for internal SRAM ??? */
	tmp3--;
	if (tmp3 != 0) {
		sram_set_size(card, 0x10000);
		sram_init(card);
		card->write_dw_old(card, 0x208, tmp);

		b2c2_info("sram size = 32K");

		return 32;
	}

	if (sram_test_location(card, 0x20000, 0x18000) != 0) {
		sram_set_size(card, 0x20000);
		sram_init(card);
		card->write_dw_old(card, 0x208, tmp);

		b2c2_info("sram size = 128K");

		return 128;
	}

	if (sram_test_location(card, 0x00000, 0x10000) != 0) {
		sram_set_size(card, 0x00000);
		sram_init(card);
		card->write_dw_old(card, 0x208, tmp);

		b2c2_info("sram size = 64K");

		return 64;
	}

	if (sram_test_location(card, 0x10000, 0x00000) != 0) {
		sram_set_size(card, 0x10000);
		sram_init(card);
		card->write_dw_old(card, 0x208, tmp);

		b2c2_info("sram size = 32K");

		return 32;
	}

	sram_set_size(card, 0x10000);
	sram_init(card);
	card->write_dw_old(card, 0x208, tmp);

	b2c2_err("SRAM detection failed. Set to 32K");

	return 0;
}

static void sll_detect_sram_size(struct b2c2_device *card)
{
	sram_detect_for_flex2(card);
}

static int eeprom_read(struct b2c2_device *card, u16 addr, u8 *buf, u16 len)
{// don't think that is correct for usb
	return card->i2c_read(card, 0x20, 0x50, addr, buf, len);
}

static u8 calc_lrc(u8 *buf, int len)
{
	int i;
	u8 sum;

	sum = 0;

	for (i = 0; i < len; i++)
		sum = sum ^ buf[i];

	return sum;
}

static int eeprom_lrc_read(struct b2c2_device *card, u32 addr, u32 len, u8 *buf, int retries)
{
	int i;

	for (i = 0; i < retries; i++) {
		if (eeprom_read(card, addr, buf, len) == len) {
			if (calc_lrc(buf, len - 1) == buf[len - 1])
				return 1;
		}
	}

	return 0;
}


static int eeprom_get_mac_addr(struct b2c2_device *card, char type, u8 *mac)
{
	u8 tmp[8];

	if (eeprom_lrc_read(card, 0x3f8, 8, tmp, 4) != 0) {
		if (type != 0) {
			mac[0] = tmp[0];
			mac[1] = tmp[1];
			mac[2] = tmp[2];
			mac[3] = 0xfe;
			mac[4] = 0xff;
			mac[5] = tmp[3];
			mac[6] = tmp[4];
			mac[7] = tmp[5];

		} else {

			mac[0] = tmp[0];
			mac[1] = tmp[1];
			mac[2] = tmp[2];
			mac[3] = tmp[3];
			mac[4] = tmp[4];
			mac[5] = tmp[5];
		}

		return 1;

	} else {

		if (type == 0) {
			memset(mac, 0, 6);

		} else {

			memset(mac, 0, 8);
		}

		return 0;
	}
}

/* PID filter */

/* every flexcop has 6 "lower" hw PID filters     */
/* these are enabled by setting bits 0-5 of 0x208 */
/* for the 32 additional filters we have to select one */
/* of them through 0x310 and modify through 0x314 */
/* op: 0=disable, 1=enable */
static void filter_enable_hw_filter(struct b2c2_device *card, int id, u8 op)
{
	b2c2_dinfo(card, "id=%d op=%d", id, op);
	if (id <= 5) {
		u32 mask = (0x00000001 << id);
		b2c2_write_reg_bitfield(card, 0x208, mask, op ? mask : 0);
	} else {
		/* select */
		b2c2_write_reg_bitfield(card, 0x310, 0x1f, (id - 6) & 0x1f);
		/* modify */
		b2c2_write_reg_bitfield(card, 0x314, 0x00006000, op ? 0x00004000 : 0);
	}
}

/* this sets the PID that should pass the specified filter */
static void pid_set_hw_pid(struct b2c2_device *card, int id, u16 pid)
{
	b2c2_dinfo(card, "id=%d  pid=%d",id, pid);
	if (id <= 5) {
		u32 adr = 0x300 + ((id & 6) << 1);
		int shift = (id & 1) ? 16 : 0;
		b2c2_dinfo(card, "id=%d  addr=%x %c  pid=%d", id, adr, (id & 1) ? 'h' : 'l', pid);
		b2c2_write_reg_bitfield(card, adr, (0x7fff) << shift, (pid & 0x1fff) << shift);
	} else {
		/* select */
		b2c2_write_reg_bitfield(card, 0x310, 0x1f, (id - 6) & 0x1f);
		/* modify */
		b2c2_write_reg_bitfield(card, 0x314, 0x1fff, pid & 0x1fff);
	}
}

void b2c2_filter_enable_mask_filter(struct b2c2_device *card, u32 op)
{
	b2c2_dinfo(card, "op=%x", op);

	b2c2_write_reg_bitfield(card, 0x208, 0x00000080, op ? 0x00000080 : 0);
}


static void ctrl_enable_mac(struct b2c2_device *card, u32 op)
{
	b2c2_write_reg_bitfield(card, 0x208, 0x00004000, op ? 0x00004000 : 0);
}

static int ca_set_mac_dst_addr_filter(struct b2c2_device *card, u8 *mac)
{
	u32 tmp1, tmp2;

	tmp1 = (mac[3] << 0x18) | (mac[2] << 0x10) | (mac[1] << 0x08) | mac[0];
	tmp2 = (mac[5] << 0x08) | mac[4];

	card->write_dw_old(card, 0x418, tmp1);
	card->write_dw_old(card, 0x41c, tmp2);

	return 0;
}

static void pid_set_group_pid(struct b2c2_device *card, u16 pid)
{
	u32 value;

	b2c2_dinfo(card, "pid=%x", pid);
	value = (pid & 0x3fff) | (card->read_dw_old(card, 0x30c) & 0xffff0000);
	card->write_dw_old(card, 0x30c, value);
}

static void pid_set_group_mask(struct b2c2_device *card, u16 pid)
{
	u32 value;

	b2c2_dinfo(card, "pid=%x", pid);
	value = ((pid & 0x3fff) << 0x10) | (card->read_dw_old(card, 0x30c) & 0xffff);
	card->write_dw_old(card, 0x30c, value);
}

static void init_pids(struct b2c2_device *card)
{
	int i;

	card->pid_count = 0;
	card->whole_bandwidth_count = 0;
	for (i = 0; i < card->useable_hw_filters; i++) {
		b2c2_info("setting filter %d to 0x1fff", i);
		card->hw_pids[i] = 0x1fff;
		pid_set_hw_pid(card, i, 0x1fff);
	}

	pid_set_group_pid(card, 0);
	pid_set_group_mask(card, 0x1fe0);
}

static void open_whole_bandwidth(struct b2c2_device *card)
{
	b2c2_dinfo(card, "Entry");
	pid_set_group_pid(card, 0);
	pid_set_group_mask(card, 0);
/*
	filter_enable_mask_filter(card, 1);
*/
}

static void close_whole_bandwidth(struct b2c2_device *card)
{
	b2c2_dinfo(card, "Entry");
	pid_set_group_pid(card, 0);
	pid_set_group_mask(card, 0x1fe0);
/*
	filter_enable_mask_filter(card, 1);
*/
}

static void whole_bandwidth_inc(struct b2c2_device *card)
{
	if (card->whole_bandwidth_count++ == 0)
		open_whole_bandwidth(card);
}

static void whole_bandwidth_dec(struct b2c2_device *card)
{
	if (--card->whole_bandwidth_count <= 0)
		close_whole_bandwidth(card);
}

/* The specified PID has to be let through the
   hw filters.
   We try to allocate an hardware filter and open whole
   bandwidth when allocation is impossible.
   All pids<=0x1f pass through the group filter.
   Returns 1 on success, -1 on error */
static int add_hw_pid(struct b2c2_device *card, u16 pid)
{
	int i;

	b2c2_dinfo(card, "pid=%d",pid);

	if (pid <= 0x1f)
		return 1;

	/* we can't use a filter for 0x2000, so no search */
	if (pid != 0x2000) {
		/* find an unused hardware filter */
		for (i = 0; i < card->useable_hw_filters; i++) {
			b2c2_dinfo(card, "pid=%d searching slot=%d", pid, i);
			if (card->hw_pids[i] == 0x1fff) {
				b2c2_dinfo(card, "pid=%d slot=%d", pid, i);
				card->hw_pids[i] = pid;
				pid_set_hw_pid(card, i, pid);
				filter_enable_hw_filter(card, i, 1);
				return 1;
			}
		}
	}
	/* if we have not used a filter, this pid depends on whole bandwidth */
	b2c2_dinfo(card, "pid=%d whole_bandwidth", pid);
	whole_bandwidth_inc(card);
	return 1;
}

/* returns -1 if the pid was not present in the filters */
static int remove_hw_pid(struct b2c2_device *card, u16 pid)
{
	int i;

	b2c2_dinfo(card, "pid=%d",pid);

	if (pid <= 0x1f)
		return 1;

	/* we can't use a filter for 0x2000, so no search */
	if (pid != 0x2000) {
		for (i = 0; i < card->useable_hw_filters; i++) {
			b2c2_dinfo(card, "pid=%d searching slot=%d", pid, i);
			if (card->hw_pids[i] == pid) {	/* find the pid slot */
				b2c2_dinfo(card, "pid=%d slot=%d", pid, i);
				card->hw_pids[i] = 0x1fff;
				pid_set_hw_pid(card, i, 0x1fff);
				filter_enable_hw_filter(card, i, 0);
				return 1;
			}
		}
	}
	/* if we have not used a filter, this pid depended on whole bandwith */
	b2c2_dinfo(card, "pid=%d whole_bandwidth", pid);
	whole_bandwidth_dec(card);
	return 1;
}

/* Adds a PID to the filters.
   Adding a pid more than once is possible, we keep reference counts.
   Whole stream available through pid==0x2000.
   Returns 1 on success, -1 on error */
int b2c2_add_pid(struct b2c2_device *card, u16 pid)
{
	int i;

	b2c2_dinfo(card, "pid=%d", pid);

	if (pid > 0x1ffe && pid != 0x2000)
		return -1;

	/* check if the pid is already present */
	for (i = 0; i < card->pid_count; i++)
		if (card->pid_list[i] == pid) {
			card->pid_rc[i]++;	/* increment ref counter */
			return 1;
		}

	if (card->pid_count == N_PID_SLOTS)
		return -1;	/* no more pids can be added */
	card->pid_list[card->pid_count] = pid;	/* register pid */
	card->pid_rc[card->pid_count] = 1;
	card->pid_count++;
	/* hardware setting */
	add_hw_pid(card, pid);

	return 1;
}

/* Removes a PID from the filters. */
int b2c2_remove_pid(struct b2c2_device *card, u16 pid)
{
	int i;

	b2c2_dinfo(card, "pid=%d", pid);

	if (pid > 0x1ffe && pid != 0x2000)
		return -1;

	/* check if the pid is present (it must be!) */
	for (i = 0; i < card->pid_count; i++) {
		if (card->pid_list[i] == pid) {
			card->pid_rc[i]--;
			if (card->pid_rc[i] <= 0) {
				/* remove from the list */
				card->pid_count--;
				card->pid_list[i]=card->pid_list[card->pid_count];
				card->pid_rc[i] = card->pid_rc[card->pid_count];
				/* hardware setting */
				remove_hw_pid(card, pid);
			}
			return 1;
		}
	}
	return -1;
}

static void decide_usable_hw_filters(struct b2c2_device *card)
{
	int hw_filters;
	int mod_option_hw_filters;

	/* FlexCop IIb & III have 6+32 hw filters
	 * FlexCop II has 6 hw filters, every other should have at least 6 */
	switch (card->b2c2_revision) {
		case 0x82:		/* II */
			hw_filters = 6;
			break;
		case 0xc3:		/* IIB */
			hw_filters = 6 + 32;
			break;
		case 0xc0:		/* III */
			hw_filters = 6 + 32;
			break;
		default:
			hw_filters = 6;
			break;
	}
	b2c2_info("the chip has %i hardware filters", hw_filters);

	mod_option_hw_filters = 0;
	if (card->enable_hw_filters >= 1)
		mod_option_hw_filters += 6;
	if (card->enable_hw_filters >= 2)
		mod_option_hw_filters += 32;

	if (mod_option_hw_filters >= hw_filters) {
		card->useable_hw_filters = hw_filters;
	} else {
		card->useable_hw_filters = mod_option_hw_filters;
		b2c2_info("Only %d filters will be used because of module option", mod_option_hw_filters);
	}
}

static int dvb_start_feed(struct dvb_demux_feed *dvbdmxfeed)
{
	struct dvb_demux *dvbdmx = dvbdmxfeed->demux;
	struct b2c2_device *card = dvbdmx->priv;

	b2c2_info("%s: PID=%d, type=%d\n", __FUNCTION__, dvbdmxfeed->pid, dvbdmxfeed->type);

	card->open_stream(card, dvbdmxfeed->pid);

	return 0;
}

static int dvb_stop_feed(struct dvb_demux_feed *dvbdmxfeed)
{
	struct dvb_demux *dvbdmx = dvbdmxfeed->demux;
	struct b2c2_device *card = dvbdmx->priv;

	b2c2_info("%s: PID=%d, type=%d\n", __FUNCTION__, dvbdmxfeed->pid, dvbdmxfeed->type);

	card->close_stream(card, dvbdmxfeed->pid);

	return 0;
}

static u32 b2c2_i2c_func(struct i2c_adapter *adapter)
{
/*	struct b2c2_device *tmp = i2c_get_adapdata(adapter);
	b2c2_dinfo(tmp,"flexcop_i2c_func");*/
	return I2C_FUNC_I2C;
}

static struct i2c_algorithm b2c2_algo = {
	.name			= "flexcop i2c algorithm",
	.id				= I2C_ALGO_BIT,
	.master_xfer	= b2c2_master_xfer,
	.functionality	= b2c2_i2c_func,
};

static int b2c2_init_dvb(struct b2c2_device *card)
{
	int ret;
	if ((ret = dvb_register_adapter(&card->dvb_adapter, 
			"B2C2/Technisat Air/Cable/Sky PCI/USB DVB", THIS_MODULE)) < 0) {
		b2c2_err("Error registering DVB adapter");
//		driver_halt(pdev); TODO ??
		return -ENODEV;
	}
	card->dvb_adapter->priv = card;

	sema_init(&card->i2c_sem,1);
	
	strncpy(card->i2c_adap.name, "B2C2/Technisat Air/Cable/Sky PCI/USB DVB i2c",I2C_NAME_SIZE);
#ifdef I2C_ADAP_CLASS_TV_DIGITAL
	card->i2c_adap.class = I2C_ADAP_CLASS_TV_DIGITAL;
#else
	card->i2c_adap.class = I2C_CLASS_TV_DIGITAL;
#endif
	card->i2c_adap.algo      = &b2c2_algo;
	card->i2c_adap.algo_data = NULL;
	card->i2c_adap.id        = I2C_ALGO_BIT;

	i2c_set_adapdata(&card->i2c_adap, card);
	
	if ((ret = i2c_add_adapter(&card->i2c_adap)) < 0) {
		b2c2_err("could not add i2c adapter.");
		goto err_i2c;
	}
	
	// DMX_MEMORY_BASED_FILTERING for USB TODO
	card->demux.dmx.capabilities = (DMX_TS_FILTERING | DMX_SECTION_FILTERING | DMX_MEMORY_BASED_FILTERING);
	card->demux.priv = card;
	card->demux.filternum = card->demux.feednum = N_PID_SLOTS;
	card->demux.start_feed = dvb_start_feed;
	card->demux.stop_feed = dvb_stop_feed;
	card->demux.write_to_decoder = NULL;
	
	if ((ret = dvb_dmx_init(&card->demux)) < 0) {
		b2c2_err("dvb_dmx failed: error %d",ret);
		goto err_dmx;
	}

	// what is a dmx frontend ??? TODO is to understand, pb
	card->hw_frontend.source = DMX_FRONTEND_0;

	card->dmxdev.filternum = card->demux.feednum;
	card->dmxdev.demux = &card->demux.dmx;
	card->dmxdev.capabilities = 0;
	if ((ret = dvb_dmxdev_init(&card->dmxdev, card->dvb_adapter)) < 0) {
		b2c2_err("dvb_dmxdev_init failed: error %d",ret);
		goto err_dmx_dev;
	}
		
	if ((ret = card->demux.dmx.add_frontend(&card->demux.dmx, &card->hw_frontend)) < 0) {
		b2c2_err("add hw_frontend to dmx failed: error %d",ret);
		goto err_dmx_add_hw_frontend;
	}

	card->mem_frontend.source = DMX_MEMORY_FE;
	if ((ret = card->demux.dmx.add_frontend(&card->demux.dmx, &card->mem_frontend)) < 0) {
		b2c2_err("add mem_frontend to dmx failed: error %d",ret);
		goto err_dmx_add_mem_frontend;
	}

	if ((ret = card->demux.dmx.connect_frontend(&card->demux.dmx, &card->hw_frontend)) < 0) {
		b2c2_err("connect frontend failed: error %d",ret);
		goto err_connect_frontend;
	}

	dvb_net_init(card->dvb_adapter, &card->dvbnet, &card->demux.dmx);
	goto success;

err_connect_frontend:
	card->demux.dmx.remove_frontend(&card->demux.dmx,&card->mem_frontend);
err_dmx_add_mem_frontend:
	card->demux.dmx.remove_frontend(&card->demux.dmx,&card->hw_frontend);
err_dmx_add_hw_frontend:
	dvb_dmxdev_release(&card->dmxdev);
err_dmx_dev:
	dvb_dmx_release(&card->demux);
err_dmx:
	i2c_del_adapter(&card->i2c_adap);
err_i2c:
	dvb_unregister_adapter(card->dvb_adapter);
	return ret;
	
success:
	return 0;
}

static int b2c2_exit_dvb(struct b2c2_device *card)
{
	dvb_net_release(&card->dvbnet);

	card->demux.dmx.close(&card->demux.dmx);
	card->demux.dmx.remove_frontend(&card->demux.dmx,&card->mem_frontend);
	card->demux.dmx.remove_frontend(&card->demux.dmx,&card->hw_frontend);
	dvb_dmxdev_release(&card->dmxdev);
	dvb_dmx_release(&card->demux);
	
	if (card->fe != NULL) 
		dvb_unregister_frontend(card->fe);
	
	if (card->dvb_adapter != NULL) {
		i2c_del_adapter(&card->i2c_adap); 
		dvb_unregister_adapter(card->dvb_adapter);
	}

	b2c2_dinfo(card,"deinitialized dvb stuff");
	return 0;
}

void b2c2_flexcop_reset(struct b2c2_device *card)
{
	card->read_dw_old(card, 0x208);
	card->write_dw_old(card, 0x208, 0);
	card->write_dw_old(card, 0x210, 0xb2ff);
	card->write_dw_old(card, 0x208, 0x40);
}

#endif

struct flexcop_device *flexcop_device_kmalloc(size_t bus_specific_len)
{
	void *bus;
	struct flexcop_device *fc = kmalloc(sizeof(struct flexcop_device), GFP_KERNEL);
	if (!fc) {
		err("no memory");
		return NULL;
	}
	memset(fc, 0, sizeof(struct flexcop_device));

	bus = kmalloc(bus_specific_len, GFP_KERNEL);
	if (!bus) {
		err("no memory");
		kfree(fc);
		return NULL;
	}
	memset(bus, 0, bus_specific_len);

	fc->bus_specific = bus;

	return fc;
}
EXPORT_SYMBOL(flexcop_device_kmalloc);

void flexcop_device_kfree(struct flexcop_device *fc)
{
	if (fc != NULL) 
		kfree(fc->bus_specific);
	kfree(fc);
}
EXPORT_SYMBOL(flexcop_device_kfree);

int flexcop_initialize(struct flexcop_device *fc)
{
	init_MUTEX(&fc->i2c_sem);
#if 0
	u32 tmp;

	card->b2c2_revision = card->read_dw_old(card, 0x204) >> 0x18;
	switch (card->b2c2_revision) {
		case 0x82:
			b2c2_info("FlexCopII(rev.130) chip found");
			break;
		case 0xc3:
			b2c2_info("FlexCopIIB(rev.195) chip found");
			break;
		case 0xc0:
			b2c2_info("FlexCopIII(rev.192) chip found");
			break;
		default:
			b2c2_err("The revision of the FlexCop chip on your card is %d", card->b2c2_revision);
			b2c2_err("This driver works only with FlexCopII(rev.130), FlexCopIIB(rev.195) and FlexCopIII(rev.192).");
			return -ENODEV;
	}

	decide_usable_hw_filters(card);

	init_pids(card);

	/* Not sure that the following is needed or indeed appropriate for USB */
	tmp = card->read_dw_old(card, 0x204);

	card->write_dw_old(card, 0x204, 0);
	mdelay(20);

	card->write_dw_old(card, 0x204, tmp);
	mdelay(10);

	tmp = card->read_dw_old(card, 0x308);
	card->write_dw_old(card, 0x308, 0x4000 | tmp);

	card->dw_sram_type = 0x10000;
	sll_detect_sram_size(card);

	b2c2_dinfo(card, "sram length = %d, sram type= %x",sram_length(card), card->dw_sram_type);

	b2c2_sram_set_media_dest(card, 1);
	b2c2_sram_set_net_dest(card, 1);

	b2c2_sram_set_cai_dest(card, 2);
	b2c2_sram_set_cao_dest(card, 2);

/* Not ready for this yet
	if (eeprom_get_mac_addr(card, 0, card->mac_addr) != 0) {
		b2c2_info("%s MAC address = %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x", __FUNCTION__,
			card->mac_addr[0], card->mac_addr[1], card->mac_addr[2], card->mac_addr[3],
			card->mac_addr[4], card->mac_addr[5], card->mac_addr[6], card->mac_addr[7]);
		ca_set_mac_dst_addr_filter(card, card->mac_addr);
		ctrl_enable_mac(card, 1);
	}
*/
	if (b2c2_init_dvb(card))
		return -ENODEV;
	b2c2_frontend_init(card);
#endif
	return 0;
}
EXPORT_SYMBOL(flexcop_initialize);

#define MODULE_DESC "B2C2 FlexcopII(b) and FlexcopIII digital TV receiver chipr"

void flexcop_exit(struct flexcop_device *card)
{
//	b2c2_exit_dvb(card);
}

flexcop_ibi_value ibi_zero;

static int flexcop_module_init(void)
{
	ibi_zero.raw = 0;
	info(MODULE_DESC " loaded successfully");
	return 0;
}

static void flexcop_module_cleanup(void)
{
	info(MODULE_DESC " unloaded successfully");
}

module_init(flexcop_module_init);
module_exit(flexcop_module_cleanup);

MODULE_DESCRIPTION(MODULE_DESC);
MODULE_LICENSE("GPL");
